﻿-- Average Combat Mana ------------------------------------------------------


-- localise global variables for faster access ------------------------------
local tremove = tremove
local type = type
local UnitMaxPower = UnitMaxPower
local UnitPower = UnitPower
local wipe = wipe
local GetAddOnMetadata = GetAddOnMetadata
local tonumber = tonumber
local tostring = tostring
local ceil = ceil
local floor = floor

-- declare addon ------------------------------------------------------------
local LibStub = _G.LibStub

local ACM = LibStub("AceAddon-3.0"):NewAddon("AverageCombatMana", "AceConsole-3.0", "AceEvent-3.0", "LibGraph-2.0")
local L = LibStub("AceLocale-3.0"):GetLocale("AverageCombatMana", true)

-- get version from .toc - set to Development if no version
local version = GetAddOnMetadata("AverageCombatMana", "Version")
--@alpha@
if version:match("@") then
	local MRG_Revision = tonumber(("$Revision: 5 $"):match("(%d+)"))
	version = "Development r"..tostring(MRG_Revision)
else
	version = "Alpha "..version
end
--@end-alpha@

-- add localisation to addon
ACM.L = L

-- declare the database
local db

-- local variables ----------------------------------------------------------
local allManaValues = {} -- stores all values during combat
local stopRed = 0
local startGreen = 0
local firstManaValue = 0 -- UnitPower at start of combat

-- additional libraries -----------------------------------------------------
local DataBroker = LibStub("LibDataBroker-1.1")
local Graph = LibStub("LibGraph-2.0")
local Media = LibStub("LibSharedMedia-3.0")
local BackDrop = LibStub("LibBackDrop-1.0")

local defaults = {
	profile = {
		barBorder = "Blizzard Tooltip",
		-- graphBorder = "Blizzard Dialog Gold",
		-- graphTexture = "Blizzard",
		addonEnabled = true,
		-- displayGraph = true,
		showBar = true,
		barBorderColor = {0.6, 0.6, 0.6, 1},
		messageFont = "Morpheus",
		messageFontSize = 18,
		messageFontEffects = monochrome
		manaFontSize = 12,
		manaFont = "Friz Quadrata TT",
		manaFontEffects = monochrome
		chatOutput = ["0-self"],
		barHeight = 20,
		barWidth = 210,
		posTextColor = {1, 1, 0},
		negTextColor = {1, 0.5, 0}
	}
}

function ACM:OnInitialize()
	-- register saved variables with AceDB
	db = LibStub("AceDB-3.0"):New("AverageCombatManaDB", defaults, "Default")
	db.RegisterCallback(self, "OnProfileChanged", "OnProfileChanged")
	db.RegisterCallback(self, "OnProfileCopied", "OnProfileChanged")
	db.RegisterCallback(self, "OnProfileReset", "OnProfileChanged")
	self.db = db
	self:SetEnabledState(self.db.profile.addonEnabled)
	
	self.maxPower = UnitPowerMax("player", 0)
	
	-- addon options table	
	self.options = self:OptionsTable() -- see AverageCombatManaOptions.lua
	-- add the 'Profiles' section
	self.options.args.profilesTab = LibStub("AceDBOptions-3.0"):GetOptionsTable(self.db)
	self.options.args.profilesTab.order = 50

	-- Register your options with AceConfigRegistry
	LibStub("AceConfig-3.0"):RegisterOptionsTable("AverageCombatMana", self.options)

	-- Add your options to the Blizz options window using AceConfigDialog
	self.optionsFrame = LibStub("AceConfigDialog-3.0"):AddToBlizOptions("AverageCombatMana", "AverageCombatMana")	

	-- support for LibAboutPanel
	if LibStub:GetLibrary("LibAboutPanel", true) then
		self.optionsFrame[L["About"]] = LibStub("LibAboutPanel").new("AverageCombatMana", "AverageCombatMana")
	end
	
	-- create the frame
	self.ManaDisplay = CreateFrame("Frame", nil, UIParent)
	self.ManaDisplay:SetPoint("CENTER")
	self.ManaDisplay:SetHeight(self.db.profile.barHeight)
	self.ManaDisplay:SetWidth(self.db.profile.barWidth)

	-- LibBackDrop-1.0 enhancement
	BackDrop:EnhanceBackdrop(self.ManaDisplay)
	
	self.ManaDisplay:SetBackDrop({
		bgFile = "Interface\\BUTTONS\\WHITE8X8", tile = false, tileSize = 16,
		edgeFile = self.db.profile.barBorder, edgeSize = 16,
		insets = { left = 4, right = 4, top = 4, bottom = 4 },
	})
	self.ManaDisplay:SetBackdropColor(0, 0, 0, 1)
	self.ManaDisplay:SetBackdropBorderColor(self.db.profile.barBorderColor)
	
	self.maxPower = UnitPowerMax("player", SPELL_POWER_MANA)
	self:CreateSegments()
	self.ManaDisplay:Show()
	
	--[[
	-- create real time graph
	self.RealTimeGraph = Graph:CreateGraphRealtime("TestRealtimeGraph",UIParent,"CENTER","CENTER",-90,90,150,150)
	self.RealTimeGraph:SetAutoScale(true)
	self.RealTimeGraph:SetGridSpacing(1, 10)
	]]--	

	-- add console commands
	self:RegisterChatCommand("acm", "SlashHandler")
	self:RegisterChatCommand("averagecombatmana", "SlashHandler")	
	
	-- create DataBroker Launcher
	if DataBroker then
		local launcher = DataBroker:NewDataObject("AverageCombatMana", {
			type = "launcher",
			icon = "Interface\\Icons\\INV_Potion_16",
			OnClick = function(clickedframe, button)
				if button == "LeftButton" then
					-- keep our options table in sync with the ldb object state
					--[[
					self.db.profile.displayGraph = not self.db.profile.displayGraph
					if self.db.profile.displayGraph then
						-- display graph code here
					else
						-- hide graph code here
					end
					LibStub("AceConfigRegistry-3.0"):NotifyChange("AverageCombatMana")
					]]--
				elseif button == "RightButton" then
					_G.InterfaceOptionsFrame_OpenToCategory(self.optionsFrame)
				--[[
				elseif button == "MiddleButton" then
					self:StartTest()
				]]--
				end
			end,
			OnTooltipShow = function(self)
			GameTooltip:AddLine("AverageCombatMana".." "..version, HIGHLIGHT_FONT_COLOR.r, HIGHLIGHT_FONT_COLOR.g, HIGHLIGHT_FONT_COLOR.b)
			-- GameTooltip:AddLine(L["Left click to display real time graph."], NORMAL_FONT_COLOR.r, NORMAL_FONT_COLOR.g, NORMAL_FONT_COLOR.b)
			-- GameTooltip:AddLine(L["Middle click to run test display."], NORMAL_FONT_COLOR.r, NORMAL_FONT_COLOR.g, NORMAL_FONT_COLOR.b)
			GameTooltip:AddLine(L["Right click for configuration."], NORMAL_FONT_COLOR.r, NORMAL_FONT_COLOR.g, NORMAL_FONT_COLOR.b)
			GameTooltip:Show()			
		end})
		self.launcher = launcher
	end
end

function ACM:OnEnable()
	self:RegisterEvent("UNIT_POWER")
	self:RegisterEvent("PLAYER_REGEN_ENABLED", "NoCombat")
	self:RegisterEvent("PLAYER_REGEN_DISABLED", "Combat")
	Media.RegisterCallback(self, "OnValueChanged", "UpdateMedia")
	if not self.ManaDisplay:IsShown() then self.ManaDisplay:Show() end
end

function ACM:OnDisable()
	self:UnregisterAllEvents()
	self.ManaDisplay:Hide()
	self.ManaDisplay:StopMovingOrSizing()
	-- self.RealTimeGraph:ResetData()
	Media.UnregisterAllCallbacks(self)
end

function ACM:OnProfileChanged()
	db = self.db
end

function ACM:SlashHandler(input)
	_G.InterfaceOptionsFrame_OpenToCategory(self.optionsFrame)
end

function ACM:UpdateMedia(callback, type, handle) 
	if type == "border" then 
		self.ManaDisplay:SetBackdrop({ 
			edgeFile = Media:Fetch("border", self.db.profile.barBorder), 
			tile = false, 
			edgeSize = 16, 
			insets = { left = 4, right = 4, top = 4, bottom = 4 }, 
		}) 
	elseif type == "font" then 
		local ManaDisplay, profile, fontStyle = self.ManaDisplay, self.db.profile

		if profile.messageFontEffects == "monochrome" or profile.manaFontEffects == "monochrome" then 
			fontStyle = "MONOCHROME" 
		elseif profile.messageFontEffects == "outline" or profile.manaFontEffects == "outline" then 
			fontStyle = "OUTLINE" 
		else 
			fontStyle = "THICKOUTLINE" 
		end 

		ManaDisplay.message:SetFont(profile.messageFont, profile.messageFontSize, fontStyle) 
		ManaDisplay.leftSide:SetFont(profile.manaFont, profile.manaFontSize, fontStyle) 
		ManaDisplay.rightSide:SetFont(profile.manaFont, profile.manaFontSize, fontStyle) 
	end 
end

function ACM:UNIT_POWER(unit, powerType)
	if (unit ~= "player") or (powerType ~= "MANA") then return end
	local power = UnitPowerMax("player")	
	if power == self.maxPower then return end
	self.maxPower = power
	self:CreateSegments()
end

local in_combat
function ACM:NoCombat()
	in_combat = false
	firstManaValue = 0
	
	if not self.db.profile.showBar then
		self.ManaDisplay:Hide()
	end
	-- reset indicator to posistion 0
	self.ManaDisplay.indicator:SetPoint("CENTER", self.ManaDisplay.segments[0], "RIGHT")
		
	-- calculate overall average gain or loss of Mana
	local allManaSum = 0
	
	local t = allManaValues
	for k = 1, #t do
		local v = t[k]
		allManaSum = allManaSum + v
	end
	local combatAverage = ceil(allManaSum / #t)
	
	self.db.char.logs = {
		events = {
			name = GetDate(),
			data = allManaValues
		}
	}
	
	if not self.db.profile.chatOutput == "2-none" then
		local chatType = self.db.profile.chatOutput:upper()
		if chatType == "0-SMARTGROUP" then
			if UnitInRaid("player") then
				chatType = "RAID"
			elseif GetNumPartyMembers() > 0 then
				chatType = "PARTY"
			end
		end
		if self.db.profile.chatOutput == "1-SELF" then
			self:Print(L["Average Mana gain or loss for the entire fight was %d"], combatAverage)
		else
			SendChatMessage(L["Average Mana gain or loss for the entire fight was %d"], chatType, nil, nil):format(combatAverage)
		end
	end
	
	wipe(allManaValues)
	combatAverage = nil
end

function ACM:Combat()
	in_combat = true
	firstManaValue = UnitPower("player")
	
	-- must show the display in combat, or nothing happens, including calculating/printing overall average
	if not self.ManaDisplay:IsShown() then
		self.ManaDisplay:Show()
	end
	
	-- reset real time graph
	-- self.RealTimeGraph:ResetData()	
end

function ACM:CreateSegments()
	self.ManaDisplay.segments = self.ManaDisplay.segments or {}
	local negSegments = floor(self.maxPower * -0.2)
	local posSegments = ceil(self.maxPower * 0.2)

	for i = negSegments, posSegments do
		local t = self.ManaDisplay.segments[i] or self.ManaDisplay:CreateTexture(nil, "ARTWORK")
		t:SetPoint("LEFT", self.ManaDisplay.segments[i - 1], "RIGHT", 0, 0)
		t:SetWidth(1)
		t:SetHeight(self.ManaDisplay:GetHeight() - 10)
		t:SetTexture(TEXTURE)
		self.ManaDisplay.segments[i] = t
	end

	-- colour the segments
	stopRed = negSegments * totalNegs * 0.1
	startGreen = posSegments * 0.1

	-- far left to -10% of the segments
	for i = negSegments, stopRed do
		self.ManaDisplay.segments[i] = self.ManaDisplay.segments[i] or self.ManaDisplay.segments[i]:SetBackdropGradient("HORIZONTAL", 1, 0, 0, 0.1, 0, 0) -- red to light red
	end
	-- -10% to 0 of the segments
	for i = stopRed + 1, 0 do
		self.ManaDisplay.segments[i] = self.ManaDisplay.segments[i] or self.ManaDisplay.segments[i]:SetBackdropGradient("HORIZONTAL", 0, 0.1, 0, 0, 1, 0) -- light blue to blue
	end
	-- 1 to +10% of the segments
	for i = 1, startGreen - 1 do
		self.ManaDisplay.segments[i] = self.ManaDisplay.segments[i] or self.ManaDisplay.segments[i]:SetBackdropGradient("HORIZONTAL", 0, 1, 0, 0, 0.1, 0) -- blue to light blue
	end
	-- +10% to far right
	for i = startGreen, posSegments do
		self.ManaDisplay.segments[i] = self.ManaDisplay.segments[i] or self.ManaDisplay.segments[i]:SetBackdropGradient("HORIZONTAL", 0, 0, 0.1, 0, 0, 1) -- light green to green
	end
	
	local fontStyle
	if self.db.profile.messageFontEffects or self.db.profile.manaFontEffects == "monochrome" then
		fontStyle = "MONOCHROME"
	elseif self.db.profile.messageFontEffects or self.db.profile.manaFontEffects == "outline" then
		fontStyle = "OUTLINE"
	else
		fontStyle = "THICKOUTLINE"
	end

	self.ManaDisplay.message = self.ManaDisplay.message or self.ManaDisplay:CreateFontString(nil, "OVERLAY")
	self.ManaDisplay.message:SetPoint("BOTTOM", self.ManaDisplay, "TOP", 0, 5)
	self.ManaDisplay.message:SetFont(self.db.profile.messageFont, self.db.profile.messageFontSize, fontStyle)
	self.ManaDisplay.message:SetText("Message Here")

	self.ManaDisplay.leftSide = self.ManaDisplay.leftSide or self.ManaDisplay:CreateFontString(nil, "OVERLAY")
	self.ManaDisplay.leftSide:SetPoint("TOPLEFT", self.ManaDisplay, "BOTTOMLEFT", 5, -5)
	self.ManaDisplay.leftSide:SetFont(self.db.profile.manaFont, self.db.profile.manaFontSize, fontStyle)
	self.ManaDisplay.leftSide:SetText(L["Full Burn!"])

	self.ManaDisplay.rightSide = self.ManaDisplay.rightSide or self.ManaDisplay:CreateFontString(nil, "OVERLAY")
	self.ManaDisplay.rightSide:SetPoint("TOPRIGHT", self.ManaDisplay, "BOTTOMRIGHT", -5, -5)
	self.ManaDisplay.rightSide:SetFont(self.db.profile.manaFont, self.db.profile.manaFontSize, fontStyle)
	self.ManaDisplay.rightSide:SetText(L["Gaining Mana!"])

	self.ManaDisplay.indicator = self.ManaDisplay.indicator or self.ManaDisplay:CreateTexture(nil, "OVERLAY")
	self.ManaDisplay.indicator:SetPoint("CENTER", self.ManaDisplay.segments[0], "RIGHT")
	self.ManaDisplay.indicator:SetWidth(10)
	self.ManaDisplay.indicator:SetHeight(self.ManaDisplay:GetHeight() + 10)
	self.ManaDisplay.indicator:SetTexture("Interface\\WorldStateFrame\\WorldState-CaptureBar")
	self.ManaDisplay.indicator:SetTexCoord(0.8, 0.8, 0, 0.3)
	self.ManaDisplay.indicator:SetDesaturated(true)
end

ACM.ManaDisplay:EnableMouse(true)
ACM.ManaDisplay:SetMovable(true)
ACM.ManaDisplay:SetClampedToScreen(true)
ACM.ManaDisplay:RegisterForDrag("LeftButton")
ACM.ManaDisplay:SetUserPlaced(false)
ACM.ManaDisplay:SetScript("OnDragStart", function(self)
	if IsAltKeyDown() then
		self.ManaDisplay:StartMoving()
	end
end)
ACM.ManaDisplay:SetScript("OnDragStop", function(self)
	self:StopMovingOrSizing()
	self:SetUserPlaced(false)
	local point, _, _, x, y = self:GetPoint(1)
	self.db.profile.barPoint = point
	self.db.profile.barX = x
	self.db.profile.barY = y
end)

-- get the 5 most recent UnitPower() values
-- or wait until we actually get 5 of them
-- calculate and return averageMana
-- the value of firstManaValue is retrieved when combat starts
local counter = 0
function ACM.ManaDisplay:SetScript("OnUpdate", function(self, elapsed) 
	counter = counter + elapsed 
	if counter < 2 then return end
	counter = 0

	local currentMana = UnitPower("player")
	local deltaMana = currentMana - firstManaValue
	firstManaValue = currentMana

	local numManaValues = #allManaValues + 1
	allManaValues[numManaValues] = deltaMana

	local count = numManaValues >= 5 and 5 or numManaValues
	for index = numManaValues - count + 1, numManaValues - 1 do
		deltaMana = deltaMana + allManaValues[index]
	end
	local averageMana = floor(deltaMana / count)
	
	self.indicator:SetPoint("CENTER", self.segments[averageMana], "RIGHT")
	
	if averageMana <= stopRed then
		self.message:SetFormattedText(L["Full Burn!\n%d"], averageMana)
		self.message:SetTextColor(self.db.profile.negTextColor)
	elseif averageMana >= startGreen then
		self.message:SetFormattedText(L["Gaining Mana!\n%d", averageMana)
		self.message:SetTextColor(self.db.profile.posTextColor)
	else
		self.message:SetText("")
	end
end)